"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Optional, Union
import asyncio
import logging
import time

import aiohttp
import requests

from toontown.exceptions import *


__all__ = ['BaseHTTPClient', 'BaseAsyncHTTPClient', 'BaseSyncHTTPClient']


Session = Union[aiohttp.ClientSession, requests.Session]


logger = logging.getLogger(__name__)


class BaseHTTPClient(ABC):
    BASE: ClassVar[str]
    HEADERS: ClassVar[dict[str, str]] = {
        'Content-Type': 'application/json',
        'User-Agent': 'toontown.py (https://github.com/jaczerob/toontown.py)'
    }

    @abstractmethod
    def __init__(self, *, session: Optional[Session] = None) -> None:
        self._session = session

    @abstractmethod
    def connect(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def request(self, method: str, path: str, headers: Optional[dict[str, str]] = None, **params: Any) -> Any: ...


class BaseSyncHTTPClient(BaseHTTPClient):
    def __init__(self, *, session: Optional[requests.Session] = None) -> None:
        self._session: requests.Session = session
        self._is_closed = True

    @property
    def is_closed(self) -> bool:
        return self._is_closed

    @is_closed.setter
    def is_closed(self, value: bool) -> None:
        self._is_closed = bool(value)

    def request(self, method: str, path: str, headers: Optional[dict[str, str]] = None, **params: Any) -> Any:
        if self.is_closed:
            raise SessionNotConnected

        url = self.BASE + path
        headers = headers or self.HEADERS

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                with self._session.request(method, url, params=params, headers=headers) as response:
                    response.raise_for_status()

                    data = response.json()
                    status = response.status_code

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        time.sleep(1 + tries * 2)
                        continue

            except requests.exceptions.HTTPError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    time.sleep(1 + tries * 2)
                    continue

                logger.info('Exhausted attempts for {0} request: {1}'.format(method, url))
                raise


class BaseAsyncHTTPClient(BaseHTTPClient):
    def __init__(self, *, session: Optional[aiohttp.ClientSession] = None) -> None:
        self._session: aiohttp.ClientSession = session

    async def connect(self) -> None:
        if self._session is None:
            self._session = aiohttp.ClientSession(raise_for_status=True)

    async def close(self) -> None:
        await self._session.close()

    async def request(self, method: str, path: str, headers: Optional[dict[str, str]] = None, **params: Any) -> Any:
        if self._session.closed:
            raise SessionNotConnected

        url = self.BASE + path
        headers = headers or self.HEADERS

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                async with self._session.request(method, url, params=params, headers=headers) as response:
                    data = await response.json()
                    status = response.status

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        await asyncio.sleep(1 + tries * 2)
                        continue

            except aiohttp.ClientResponseError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)
                raise

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    time.sleep(1 + tries * 2)
                    continue

                logger.info('Exhausted attempts for {0} request: {1}'.format(method, url))
                raise



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from pathlib import Path
from typing import Optional, Union
import aiohttp
import requests

from ..http import RewrittenSyncHTTPClient, RewrittenAsyncHTTPClient
from ..models.rewritten import *
from .base import BaseAsyncToontownClient, BaseSyncToontownClient


__all__ = ['RewrittenSyncToontownClient', 'RewrittenAsyncToontownClient']


class RewrittenSyncToontownClient(BaseSyncToontownClient):
    """Synchronous client to interact with the Toontown Rewritten API"""

    def __init__(self, *, session: Optional[requests.Session] = None) -> None:
        super().__init__(RewrittenSyncHTTPClient(session=session))

    def status(self) -> Status:
        data = self.http.request(
            'GET',
            '/status'
        )

        return Status(**data)

    def release_notes(self, *, id: Optional[int] = None) -> ReleaseNotesList:
        if id is not None:
            path = f'/releasenotes/{id}'
        else:
            path = '/releasenotes'

        data = self.http.request(
            'GET',
            path
        )

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return ReleaseNotesList(**data)
        
    def news(self, *, id: Optional[int] = None, all: Optional[bool] = False) -> NewsList:
        if id is not None:
            path = f'/news/{id}'
        elif all is True:
            path = '/news/list'
        else:
            path = '/news'

        data = self.http.request(
            'GET',
            path
        )

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return NewsList(**data)

    def doodles(self) -> Doodles:
        data = self.http.request(
            'GET',
            '/doodles'
        )

        return Doodles(**data)

    def field_offices(self) -> FieldOffices:
        data = self.http.request(
            'GET',
            '/fieldoffices'
        )

        return FieldOffices(**data)

    def invasions(self) -> Invasions:
        data = self.http.request(
            'GET',
            '/invasions'
        )

        return Invasions(**data)

    def login(
        self, 
        *, 
        username: Optional[str] = None, 
        password: Optional[str] = None,
        app_token: Optional[str] = None,
        auth_token: Optional[str] = None,
        queue_token: Optional[str] = None,
    ) -> Login:
        params = {'format': 'json'}

        if app_token is not None and auth_token is not None:
            params['appToken'] = app_token
            params['authToken'] = auth_token
        elif queue_token is not None:
            params['queueToken'] = queue_token
        elif username is not None and password is not None:
            params['username'] = username
            params['password'] = password
        else:
            raise Exception('Please provide either a username and password, a queue token, or a auth token and app token to log in')

        headers = self.http.HEADERS | {'Content-Type': 'application/x-www-form-urlencoded'}

        data = self.http.request(
            'POST',
            '/login',
            headers=headers,
            **params
        )
        
        return Login(**data)

    def population(self) -> Population:
        data = self.http.request(
            'GET',
            '/population'
        )

        return Population(**data)

    def silly_meter(self) -> None:
        data = self.http.request(
            'GET',
            '/sillymeter'
        )

        return SillyMeter(**data)

    def update(self, path: Union[str, Path]) -> None:
        self.http.update(path)


class RewrittenAsyncToontownClient(BaseAsyncToontownClient):
    """Asynchronous client to interact with the Toontown Rewritten API"""

    def __init__(self, *, session: Optional[aiohttp.ClientSession] = None) -> None:
        super().__init__(RewrittenAsyncHTTPClient(session=session))

    async def connect(self) -> None:
        await self.http.connect()

    async def close(self) -> None:
        await self.http.close()

    async def status(self) -> Status:
        data = await self.http.request(
            'GET',
            '/status'
        )

        return Status(**data)

    async def release_notes(self, *, id: Optional[int] = None) -> ReleaseNotesList:
        if id is not None:
            path = f'/releasenotes/{id}'
        else:
            path = '/releasenotes'

        data = await self.http.request(
            'GET',
            path
        )

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return ReleaseNotesList(**data)
        
    async def news(self, *, id: Optional[int] = None, all: Optional[bool] = False) -> NewsList:
        if id is not None:
            path = f'/news/{id}'
        elif all is True:
            path = '/news/list'
        else:
            path = '/news'

        data = await self.http.request(
            'GET',
            path
        )

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return NewsList(**data)

    async def doodles(self) -> Doodles:
        data = await self.http.request(
            'GET',
            '/doodles'
        )

        return Doodles(**data)

    async def field_offices(self) -> FieldOffices:
        data = await self.http.request(
            'GET',
            '/fieldoffices'
        )

        return FieldOffices(**data)

    async def invasions(self) -> Invasions:
        data = await self.http.request(
            'GET',
            '/invasions'
        )

        return Invasions(**data)

    async def login(
        self, 
        *, 
        username: Optional[str] = None, 
        password: Optional[str] = None,
        app_token: Optional[str] = None,
        auth_token: Optional[str] = None,
        queue_token: Optional[str] = None,
    ) -> Login:
        params = {'format': 'json'}

        if app_token is not None and auth_token is not None:
            params['appToken'] = app_token
            params['authToken'] = auth_token
        elif queue_token is not None:
            params['queueToken'] = queue_token
        elif username is not None and password is not None:
            params['username'] = username
            params['password'] = password
        else:
            raise Exception('Please provide either a username and password, a queue token, or a auth token and app token to log in')

        headers = self.http.HEADERS | {'Content-Type': 'application/x-www-form-urlencoded'}

        data = await self.http.request(
            'POST',
            '/login',
            headers=headers,
            **params
        )
        
        return Login(**data)

    async def population(self) -> Population:
        data = await self.http.request(
            'GET',
            '/population'
        )

        return Population(**data)

    async def silly_meter(self) -> SillyMeter:
        data = await self.http.request(
            'GET',
            '/sillymeter'
        )

        return SillyMeter(**data)

    async def update(self, path: Union[str, Path]) -> None:
        await self.http.update(path)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from typing import Literal, Optional


__all__ = ['Login']


@dataclass
class Login:
    """"Wrapper class for /login response
    
    Attributes
    ----------
    success : Literal['false', 'partial', 'true', 'delayed']
        what type of response the login gave

    banner : Optional[str]
        an optional reason from the server if `false` or `partial` was given for success

    response_token : Optional[str]
        if success is `partial`, a token to verify a ToonGuard response to the server

    gameserver : Optional[str]
        if success is `true`, Toontown Rewritten's gameserver IP

    cookie : Optional[str]
        if success is `true`, your Toontown Rewritten session identifier

    eta : Optional[int]
        if success is `delayed`, how many seconds you are estimated to be able to play

    position : Optional[int]
        if success is `delayed`, how many toons are ahead of you in queue

    queue_token : Optional[str]
        if success is `delayed`, this identifier holds your position in queue
    """

    __slots__ = ['success', 'banner', 'response_token', 'gameserver', 'cookie', 'eta', 'position', 'queue_token']

    def __init__(self, **payload) -> None:
        self.success: Literal['false', 'partial', 'true', 'delayed'] = payload.get('success')
        self.banner: Optional[str] = payload.get('banner', None)
        self.response_token: Optional[str] = payload.get('responseToken', None)
        self.gameserver: Optional[str] = payload.get('gameserver', None)
        self.cookie: Optional[str] = payload.get('cookie', None)
        self.eta: Optional[int] = payload.get('eta', None)
        self.position: Optional[int] = payload.get('position', None)
        self.queue_token: Optional[str] = payload.get('queueToken', None)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass

from ..base import BaseAPIModel


__all__ = ['Doodle', 'Doodles']


@dataclass
class Doodle:
    """Wrapper class for doodle data

    Attributes
    ----------
    district : str
        the district the doodle is in

    playground : str
        the playground the doodle is in

    dna : str
        the doodle's DNA string

    rendition : str
        a link to a 256x256 png of the doodle's rendition

    traits : list[str]
        the list of the doodle's traits

    cost : int
        how much the doodle costs to purchase
    """

    __slots__ = ['district', 'playground', 'dna', 'traits', 'cost']

    district: str
    playground: str
    dna: str
    traits: list[str]
    cost: int

    @property
    def rendition(self) -> str:
        """A link to a 256x256 png of the doodle's rendition"""
        return f'https://rendition.toontownrewritten.com/render/{self.dna}/doodle/256x256.png'


class Doodles(BaseAPIModel[Doodle]):
    """Wrapper class for the /doodles response
    
    A tuple-like class containing `Doodle` objects, sorted by district and playground
    """
    
    def __init__(self, **payload) -> None:
        iterable = tuple(
            sorted([
                Doodle(district, playground, **doodle)
                for district, playgrounds in payload.items()
                for playground, doodles in playgrounds.items()
                for doodle in doodles
            ], key = lambda doodle: (doodle.district, doodle.playground))
        )

        super().__init__(iterable)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from ..base import BaseAPIModel


__all__ = ['FieldOffice', 'FieldOffices']


ZONE_LOOKUP = {
    '3100': 'Walrus Way',
    '3200': 'Sleet Street',
    '3300': 'Polar Place',
    '4100': 'Alto Avenue',
    '4200': 'Baritone Boulevard',
    '4300': 'Tenor Terrace',
    '5100': 'Elm Street',
    '5200': 'Maple Street',
    '5300': 'Oak Street',
    '9100': 'Lullaby Lane',
    '9200': 'Pajama Place',
}

DEPARTMENT_LOOKUP = {
    's': 'Sellbot'
}


@dataclass
class FieldOffice:
    """Wrapper class for field office data
    
    Attributes
    ----------
    last_updated : datetime
        when the field office's data was last updated

    street : str
        the street that the field office is on

    department : str
        the department of the field office (currently only `Sellbot`)

    difficulty : int
        how many stars the field office is

    annexes : int
        how many annexes are left in the field office

    open : bool
        whether or not the elevator to the field office is open

    expiring : Optional[datetime]
        shows how long left the field office will stand after the last annex is defeated, otherwise `None`
    """

    __slots__ = ['last_updated', 'street', 'department', 'difficulty',
                 'annexes', 'open', 'expiring']
    
    def __init__(self, last_updated, zone, *, department, difficulty, annexes, open, expiring) -> None:
        self.last_updated: datetime = last_updated
        self.street: str = ZONE_LOOKUP[zone]
        self.department: str = DEPARTMENT_LOOKUP[department]
        self.difficulty: int = difficulty + 1
        self.annexes: int = annexes
        self.open: bool = open
        self.expiring: Optional[datetime] = datetime.fromtimestamp(expiring) if expiring else None


class FieldOffices(BaseAPIModel[FieldOffice]):
    """"Wrapper class for /fieldoffices response

    A tuple-like class containing `FieldOffice` objects, sorted by difficulty

    Attributes
    ----------
    last_updated : datetime
        the time when the field offices data was last updated
    """

    __slots__ = ['last_updated']
    
    def __init__(self, **payload) -> None:
        self.last_updated = last_updated = datetime.fromtimestamp(payload.pop('lastUpdated'))
        iterable = tuple(
            sorted([
                FieldOffice(last_updated, zone, **props)
                for zone, props in payload.pop('fieldOffices').items()
            ],
            key=lambda field_office: field_office.difficulty,
            reverse=True)
        )

        super().__init__(iterable)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from abc import ABC, abstractmethod

from ..http import BaseHTTPClient


__all__ = ['BaseToontownClient', 'BaseSyncToontownClient', 'BaseAsyncToontownClient']


class BaseToontownClient(ABC):
    """Base Toontown Client to provide functionality for users to interact with the Toontown Rewritten API"""

    @abstractmethod
    def __init__(self, httpclient: BaseHTTPClient) -> None:
        self.http = httpclient

    @abstractmethod
    def connect(self) -> None:
        """Connect to the HTTP client
        
        Must be called before using any other methods in this class
        """

    @abstractmethod
    def close(self) -> None:
        """Closes connection to the HTTP client"""


class BaseSyncToontownClient(BaseToontownClient):
    def connect(self) -> None:
        self.http.connect()

    def close(self) -> None:
        self.http.close()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.close()


class BaseAsyncToontownClient(BaseToontownClient):
    async def connect(self) -> None:
        await self.http.connect()

    async def close(self) -> None:
        await self.http.close()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args):
        await self.close()



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from datetime import datetime

from ..base import BaseAPIModel


__all__ = ['Invasion', 'Invasions']


@dataclass
class Invasion:
    """Wrapper class for invasion data
    
    Attributes
    ----------
    district : str
        the district the invasion is in

    as_of : datetime
        when the invasion started

    type : str
        the Cog type of the invasion

    progress : int
        how many cogs were defeated in this invasion

    total : int
        how many cogs that are invading

    is_mega_invasion : bool
        whether or not this is a mega invasion
    """

    __slots__ = ['district', 'as_of', 'type', 'progress', 'total', 'is_mega_invasion']

    def __init__(self, district, **payload) -> None:
        self.district: str = district
        self.as_of = datetime.fromtimestamp(payload.pop('asOf'))
        self.type: str = payload.pop('type')

        progress, total = payload.pop('progress').split('/')
        self.progress = int(progress)
        self.total = int(total)
        self.is_mega_invasion: bool = total == 1000000


class Invasions(BaseAPIModel[Invasion]):
    """"Wrapper class for /invasions response

    A tuple-like class containing `Invasion` objects

    Attributes
    ----------
    last_updated : datetime
        the time when the invasions were last updated
    """

    __slots__ = ['last_updated']
    
    def __init__(self, **payload) -> None:
        iterable = tuple(Invasion(**{'district': key} | item) for key, item in payload.pop('invasions').items())
        super().__init__(iterable)

        self.last_updated = datetime.fromtimestamp(payload.pop('lastUpdated'))
        


"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from datetime import datetime

from ..base import BaseAPIModel


__all__ = ['Population', 'District']


@dataclass
class District:
    """Wrapper class for district data
    
    Attributes
    ----------
    name : str
        the name of the district
        
    population : int
        the population of the district
        
    last_updated : datetime
        when the district's population was last updated by the server
    """
    
    name: str
    population: int
    last_updated: datetime


class Population(BaseAPIModel[District]):
    """"Wrapper class for /population response

    A tuple-like class containing `District` objects, sorted by population

    Attributes
    ----------
    total : int
        the total population of Toontown Rewritten

    last_updated : datetime
        the time of the last population update
    """

    __slots__ = ['total', 'last_updated']

    def __init__(self, **payload) -> None:
        self.total: int = payload.get('totalPopulation')
        self.last_updated = last_updated = datetime.fromtimestamp(payload.get('lastUpdated'))

        iterable = tuple(sorted([
            District(name, population, last_updated)
            for name, population in payload.get('populationByDistrict').items()
        ], key=lambda district: district.population))
        
        super().__init__(iterable)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import ClassVar, Union
import asyncio
import bz2
import hashlib
import logging
import os
import platform
import sys

from ..exceptions import *
from .base import BaseAsyncHTTPClient, BaseSyncHTTPClient


__all__ = ['RewrittenSyncHTTPClient', 'RewrittenAsyncHTTPClient']


logger = logging.getLogger(__name__)

MANIFEST = 'https://cdn.toontownrewritten.com/content/patchmanifest.txt'
PATCHES = 'https://download.toontownrewritten.com/patches'

CHUNK_SIZE = 10 * 10 * 1024


def get_platform():
    if sys.platform == 'win32' and platform.machine().endswith('64'):
        return 'win64' # for handling 64 bit files in the patch manifest
    return sys.platform


class RewrittenSyncHTTPClient(BaseSyncHTTPClient):
    BASE: ClassVar[str] = 'https://toontownrewritten.com/api'

    def update(self, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)

        if not path.is_dir():
            raise Exception('Path does not exist or is not a directory')

        manifest = self._session.get(
            MANIFEST,
            headers=self.HEADERS,
        ).json()

        pf = get_platform()

        def try_update(args):
            file, props = args
            if pf not in props['only']:
                return

            file_path = path.joinpath(file)

            if file_path.exists():
                hash = hashlib.sha1(file_path.open('rb').read()).hexdigest()
                if props['hash'] == hash:
                    return

            url = '{0}/{1}'.format(PATCHES, props['dl'])

            with self._session.get(url, stream=True) as resp:
                logger.info(f'Downloading {url} to {file_path}')
                resp.raise_for_status()
                decompressor = bz2.BZ2Decompressor()

                with open(file_path, 'wb') as file:
                    for chunk in resp.iter_content(chunk_size=CHUNK_SIZE):
                        file.write(decompressor.decompress(chunk))

        with ThreadPool(os.cpu_count()) as pool:
            pool.map(try_update, manifest.items())


class RewrittenAsyncHTTPClient(BaseAsyncHTTPClient):
    BASE: ClassVar[str] = 'https://toontownrewritten.com/api'

    async def update(self, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)

        if not path.is_dir():
            raise Exception('Path does not exist or is not a directory')

        manifest = await (await self._session.get(
            MANIFEST,
            headers=self.HEADERS,
        )).json()

        pf = get_platform()

        async def try_update(args):
            file, props = args
            if pf not in props['only']:
                return

            file_path = path.joinpath(file)

            if file_path.exists():
                hash = hashlib.sha1(file_path.open('rb').read()).hexdigest()
                if props['hash'] == hash:
                    return

            url = '{0}/{1}'.format(PATCHES, props['dl'])

            async with self._session.get(url) as resp:
                logger.info(f'Downloading {url} to {file_path}')

                resp.raise_for_status()
                decompressor = bz2.BZ2Decompressor()

                with open(file_path, 'wb') as file:
                    async for chunk in resp.content.iter_chunked(CHUNK_SIZE):
                        file.write(decompressor.decompress(chunk))

        await asyncio.gather(*map(try_update, manifest.items()))



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal, Optional

from ..base import BaseAPIModel


@dataclass
class SillyTeam:
    """Wrapper class for current rewards in the /sillymeter
    
    Attributes
    ----------
    reward : str
        the reward this Silly Team will give

    description : str
        the description of the reward this Silly Team will give

    points : int
        the amount of points this Silly Team earned in the cycle, 
        set to `None` if the Silly Meter state is not set to `Reward`
    """

    __slots__ = ['reward', 'description', 'points']

    reward: str
    description: str
    points: Optional[int]


class SillyMeter(BaseAPIModel[SillyTeam]):
    """Wrapper class for /sillymeter response

    A tuple-like class containing `SillyTeam` objects, sorted by points

    Attributes
    ----------
    state : Literal['Active', 'Reward', 'Inactive']
        the current state of the Silly Meter

    health : int
        the current health of the Silly Meter (0 - 5,000,000)

    winner : Optional[str]
        the winning Silly Team whose reward is currently active, otherwise `None`

    next_update_timestamp : datetime
        when the Silly Meter will next update itself

    as_of : datetime
        when the server generated the Silly Meter data
    """

    __slots__ = ['state', 'health', 'winner', 'next_update_timestamp', 'as_of']

    def __init__(self, **payload: dict[str, Any]) -> None:
        self.state: Literal['Active', 'Reward', 'Inactive'] = payload.get('state')
        self.health: int = payload.get('hp')

        self.winner: Optional[str] = payload.get('winner')
        self.next_update_timestamp: datetime = datetime.fromtimestamp(payload.get('nextUpdateTimestamp'))
        self.as_of: datetime = datetime.fromtimestamp(payload.get('asOf'))

        rewards: list[str] = payload.get('rewards')
        reward_descriptions: list[str] = payload.get('rewardDescriptions')
        reward_points: list[Optional[int]] = payload.get('rewardPoints')
        iterable = map(lambda args: SillyTeam(*args), zip(rewards, reward_descriptions, reward_points))

        if self.state == 'Reward':
            iterable = tuple(sorted(iterable, key=lambda silly_team: silly_team.points))
        else:
            iterable = tuple(iterable)

        super().__init__(iterable)



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
from datetime import datetime

from ..base import BaseAPIModel


__all__ = ['District', 'Districts']


@dataclass
class District:
    """Wrapper class for district data
    
    Attributes
    ----------
    name : str
        district name
    
    online : bool
        generally always true, districts are removed if they don't report their status 
        after a few minutes

    population : int
        district population

    invasion_online : bool
        true if invasion is present in district

    last_update : datetime
        the time at which the district was last updated

    cogs_attacking : str
        the cogs attacking name

    count_defeated : int
        amount of cogs defeated in the district

    remaining_time : int
        amount of time before the invasion automatically ends in seconds
    """

    __slots__ = [
        'name', 'online', 'population', 'invasion_online', 'last_update',
        'cogs_attacking', 'count_defeated', 'count_total', 'remaining_time'
    ]

    def __init__(self, **props) -> None:
        self.name: str = props.get('name')
        self.online: bool = props.get('online')
        self.population: int = props.get('population')
        self.invasion_online: bool = props.get('invasion_online')
        self.last_update: datetime = datetime.fromtimestamp(props.get('last_update'))
        self.cogs_attacking: str = props.get('cogs_attacking')
        self.count_defeated: int = props.get('count_defeated')
        self.count_total: int = props.get('count_total')
        self.remaining_time: int = props.get('remaining_time')


class Districts(BaseAPIModel[District]):
    """Wrapper class for the /districts response
    
    A tuple-like class containing `District` objects, sorted by district and playground
    """

    def __init__(self, **payload) -> None:
        iterable = tuple(
            sorted([
                District(**district)
                for district in payload['iterable']
            ], key=lambda district: district.population)
        )

        super().__init__(iterable)



import pathlib
import re
from setuptools import setup

HERE = pathlib.Path(__file__).parent
readme = (HERE / 'README.md').read_text()

packages = [
    'toontown',
    'toontown.client',
    'toontown.http',
    'toontown.models',
    'toontown.models.clash',
    'toontown.models.rewritten'
]


def requirements():
    with open('requirements.txt') as file:
        return file.read().splitlines()


def version():
    with open('toontown/__init__.py') as f:
        return re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)


setup(
    name='toontown.py',
    author='jaczerob',
    url='https://github.com/jaczerob/toontown.py',
    version=version(),
    packages=packages,
    license='MIT',
    description='A simple Python wrapper for the Toontown Rewritten API',
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=requirements(),
    python_requires='>=3.8.0',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Topic :: Games/Entertainment',
    ],
)



"""
Toontown Rewritten API Wrapper
==============================

A basic wrapper for the Toontown Rewritten/Corporate Clash API

:copyright: (c) 2022-present jaczerob
:license: MIT, see LICENSE for more details.
"""

__title__ = 'toontown'
__author__ = 'jaczerob'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present jaczerob'
__version__ = '2.0.1'

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

import logging

from .exceptions import *

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import Optional
import aiohttp
import requests

from ..http import ClashSyncHTTPClient, ClashAsyncHTTPClient
from ..models.clash import *
from .base import BaseAsyncToontownClient, BaseSyncToontownClient


__all__ = ['ClashSyncToontownClient', 'ClashAsyncToontownClient']


class ClashSyncToontownClient(BaseSyncToontownClient):
    """Synchronous client to interact with the Toontown Corporate Clash API"""

    def __init__(self, *, session: Optional[requests.Session] = None) -> None:
        super().__init__(ClashSyncHTTPClient(session=session))

    def districts(self) -> Districts:
        data = self.http.request(
            'GET',
            '/districts.js'
        )

        return Districts(**dict(iterable=data))

    def login(self, username: str, password: str) -> Login:
        data = self.http.request(
            'GET',
            f'/login/{username}',
            password=password
        )

        return Login(**data)

    def news(self, id: Optional[int] = None) -> NewsList:
        if id is not None:
            path = f'/launcher/news/{id}'
        else:
            path = '/launcher/news'

        data = self.http.request(
            'GET',
            path
        )

        return NewsList(**dict(iterable=data))


class ClashAsyncToontownClient(BaseAsyncToontownClient):
    """Asynchronous client to interact with the Toontown Corporate Clash API"""

    def __init__(self, *, session: Optional[aiohttp.ClientSession] = None) -> None:
        super().__init__(ClashAsyncHTTPClient(session=session))

    async def districts(self) -> Districts:
        data = await self.http.request(
            'GET',
            '/districts.js'
        )

        return Districts(**dict(iterable=data))

    async def login(self, username: str, password: str) -> Login:
        data = await self.http.request(
            'GET',
            f'/login/{username}',
            password=password
        )

        return Login(**data)

    async def news(self, id: Optional[int] = None) -> NewsList:
        if id is not None:
            path = f'/launcher/news/{id}'
        else:
            path = '/launcher/news'

        data = await self.http.request(
            'GET',
            path
        )

        return NewsList(**dict(iterable=data))



"""
The MIT License (MIT)

Copyright (c) 2022-present jaczerob

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from dataclasses import dataclass
import re
from datetime import datetime
from typing import  Optional

from ..base import BaseAPIModel


__all__ = ['ReleaseNotes', 'ReleaseNotesList']


DATE_FMT = '%B %-d, %Y at %-I:%M %p'

HTMLELEMENT_CLEANER = re.compile(r'(<.*?>|\r)')
HTMLENCODE_CLEANER = re.compile(r'&nbsp;')
NEWLINE_CLEANER = re.compile(r'\n{2,}')


def clean(string: Optional[str]):
    """Strips out HTML elements and cleans up new lines from post body"""
    if string is None:
        return string

    string = HTMLELEMENT_CLEANER.sub('', string)
    string = HTMLENCODE_CLEANER.sub(' ', string)
    string = NEWLINE_CLEANER.sub('\n\n', string)

    return string


@dataclass
class ReleaseNotes:
    """Wrapper class for the release note data
    
    Attributes
    ----------
    note_id : int
        the ID of the post
        
    slug : str
        the slug of the post

    date : datetime
        when the post was written

    body_raw : str
        the raw body of the post containing HTML elements

    body : str
        a readable version of the body, with HTML elements stripped out
    """

    __slots__ = ['note_id', 'slug', 'date', 'body_raw', 'body']

    def __init__(self, **props) -> None:
        self.note_id: int = props.get('noteId')
        self.slug: str = props.get('slug')
        self.date: datetime = datetime.strptime(props.get('date'), DATE_FMT)

        body_raw = props.get('body', None)
        self.body_raw: Optional[str] = body_raw
        self.body: Optional[str] = clean(body_raw)


class ReleaseNotesList(BaseAPIModel[ReleaseNotes]):
    """Wrapper class for the /releasenotes response
    
    A tuple-like class containing `ReleaseNote` objects
    """
    
    def __init__(self, **payload) -> None:
        iterable = tuple(payload.get('iterable'))
        super().__init__(iterable)
